from relevanceai.client.helpers import Credentials
from relevanceai.utils.base import _Base
from typing import List, Optional, Dict


class CentroidsClient(_Base):
    def __init__(self, credentials: Credentials):
        super().__init__(credentials)

    def list_closest_to_center(
        self,
        dataset_id: str,
        vector_fields: List,
        alias: str,
        cluster_ids: Optional[List] = None,
        centroid_vector_fields: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        facets: Optional[List] = None,
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        include_facets: bool = False,
        cluster_properties_filter: Optional[Dict] = {},
        verbose: bool = False,
    ):
        """
        List of documents closest from the center.
        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_fields: list
            The vector fields where a clustering task runs
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        centroid_vector_fields: list
            Vector fields stored
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results
        """
        cluster_ids = [] if cluster_ids is None else cluster_ids
        centroid_vector_fields = (
            [] if centroid_vector_fields is None else centroid_vector_fields
        )
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters
        facets = [] if facets is None else facets

        if not centroid_vector_fields:
            centroid_vector_fields = vector_fields
        parameters = {
            "vector_fields": vector_fields,
            "centroid_vector_fields": centroid_vector_fields,
            "alias": alias,
            "dataset_id": dataset_id,
            "cluster_ids": cluster_ids,
            "select_fields": select_fields,
            "approx": approx,
            "sum_fields": sum_fields,
            "page_size": page_size,
            "page": page,
            "similarity_metric": similarity_metric,
            "filters": filters,
            "facets": facets,
            "min_score": min_score,
            "include_vector": include_vector,
            "include_count": include_count,
            "include_facets": include_facets,
            "cluster_properties_filter": cluster_properties_filter,
        }
        endpoint = f"/datasets/{dataset_id}/cluster/centroids/list_closest_to_center"
        method = "POST"
        # self._log_to_dashboard(
        #     method=method,
        #     parameters=parameters,
        #     endpoint=endpoint,
        #     dashboard_type="cluster_centroids_closest",
        #     verbose=verbose,
        # )
        return self.make_http_request(endpoint, method=method, parameters=parameters)

    documents_closest_to_center = list_closest_to_center

    def list_furthest_from_center(
        self,
        dataset_id: str,
        vector_fields: List,
        alias: str,
        centroid_vector_fields: Optional[List] = None,
        cluster_ids: Optional[List] = None,
        select_fields: Optional[List] = None,
        approx: int = 0,
        sum_fields: bool = True,
        page_size: int = 1,
        page: int = 1,
        similarity_metric: str = "cosine",
        filters: Optional[List] = None,
        facets: Optional[List] = None,
        min_score: int = 0,
        include_vector: bool = False,
        include_count: bool = True,
        include_facets: bool = False,
        cluster_properties_filter: Optional[Dict] = {},
    ):
        """
        List of documents furthest from the center.
        Parameters
        ----------
        dataset_id: string
            Unique name of dataset
        vector_fields: list
            The vector field where a clustering task was run.
        cluster_ids: list
            Any of the cluster ids
        alias: string
            Alias is used to name a cluster
        select_fields: list
            Fields to include in the search results, empty array/list means all fields
        approx: int
            Used for approximate search to speed up search. The higher the number, faster the search but potentially less accurate
        sum_fields: bool
            Whether to sum the multiple vectors similarity search score as 1 or seperate
        page_size: int
            Size of each page of results
        page: int
            Page of the results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        filters: list
            Query for filtering the search results
        facets: list
            Fields to include in the facets, if [] then all
        min_score: int
            Minimum score for similarity metric
        include_vectors: bool
            Include vectors in the search results
        include_count: bool
            Include the total count of results in the search results
        include_facets: bool
            Include facets in the search results
        cluster_properties_filter: dict
            Filter if clusters with certain characteristics should be hidden in results
        """
        centroid_vector_fields = (
            [] if centroid_vector_fields is None else centroid_vector_fields
        )
        cluster_ids = [] if cluster_ids is None else cluster_ids
        select_fields = [] if select_fields is None else select_fields
        filters = [] if filters is None else filters
        facets = [] if facets is None else facets

        if not centroid_vector_fields:
            centroid_vector_fields = vector_fields
        endpoint = f"/datasets/{dataset_id}/cluster/centroids/list_furthest_from_center"
        method = "POST"
        parameters = {
            "dataset_id": dataset_id,
            "vector_fields": vector_fields,
            "alias": alias,
            "cluster_ids": cluster_ids,
            "select_fields": select_fields,
            "centroid_vector_fields": centroid_vector_fields,
            "approx": approx,
            "sum_fields": sum_fields,
            "page_size": page_size,
            "page": page,
            "similarity_metric": similarity_metric,
            "filters": filters,
            "facets": facets,
            "min_score": min_score,
            "include_vector": include_vector,
            "include_count": include_count,
            "include_facets": include_facets,
            "cluster_properties_filter": cluster_properties_filter,
        }
        # self._log_to_dashboard(
        #     method=method,
        #     parameters=parameters,
        #     endpoint=endpoint,
        #     dashboard_type="cluster_centroids_furthest",
        # )
        response = self.make_http_request(
            endpoint, method=method, parameters=parameters
        )
        return response

    documents_furthest_from_center = list_furthest_from_center

    def update(
        self,
        dataset_id: str,
        vector_fields: List[str],
        alias: str,
        cluster_centers: List[Dict[str, List[float]]],
        centroid_vector_fields: List[str] = None,
    ):
        """
        API reference link: https://api.us-east-1.relevance.ai/latest/core/documentation#operation/UpdateClusterCentroids

        Update the centroids contained within your dataset

        Parameters
        ----------

        dataset_id: str
            The name of the dataset

        vector_fields: List[str]
            A list of the vectors fields in your dataset that have cluster centroids you wish to update

        alias: str
            The alias that was used to cluster

        cluster_centers: List[Dict[str: List[float]]]
            A List containing dictionaries of cluster id's to be updated, with their keys being the new centroids
        """
        if centroid_vector_fields is None:
            centroid_vector_fields = []
        return self.make_http_request(
            endpoint=f"/datasets/{dataset_id}/cluster/centroids/update",
            method="POST",
            parameters={
                "vector_fields": vector_fields,
                "centroid_vector_fields": centroid_vector_fields,
                "alias": alias,
                "cluster_centers": cluster_centers,
            },
        )

    def delete_centroid_by_id(
        self, centroid_id: str, dataset_id: str, vector_field: str, alias: str
    ):
        """
        OLD API reference link: https://api.us-east-1.relevance.ai/latest/documentation#operation/delete_centroids_api_services_cluster_centroids__centroid_id__delete_post

        Delete a centroid by ID

        Parameters
        ----------

        centroid_id: str
            The id of the centroid

        dataset_id: str
            The name of the dataset

        vector_field: str
            The vector_field that contains the cluster id

        alias: str
            The alias that was used to cluster
        """

        return self.make_http_request(
            endpoint=f"services/cluster/centroids/{centroid_id}/delete",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "vector_field": vector_field,
                "alias": alias,
            },
        )

    def insert(
        self,
        dataset_id: str,
        cluster_centers: List,
        vector_fields: List,
        alias: str = "default",
    ):
        """
        Insert your own cluster centroids for it to be used in approximate search settings and cluster aggregations.
        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        cluster_centers : list
            Cluster centers with the key being the index number
        vector_field: string
            The vector field where a clustering task was run.
        alias: string
            Alias is used to name a cluster
        """
        return self.make_http_request(
            f"/datasets/{dataset_id}/cluster/centroids/insert",
            method="POST",
            parameters={
                "dataset_id": dataset_id,
                "cluster_centers": cluster_centers,
                "vector_fields": vector_fields,
                "alias": alias,
            },
        )

    def documents(
        self,
        dataset_id: str,
        vector_fields: List,
        cluster_ids: Optional[List] = None,
        alias: str = "default",
        page_size: int = 5,
        cursor: str = None,
        page: int = 1,
        include_vector: bool = False,
        similarity_metric: str = "cosine",
    ):
        """
        Retrieve the cluster centroids by IDs

        Parameters
        -------------

        dataset_id : string
            Unique name of dataset
        cluster_ids : list
            List of cluster IDs
        vector_fields: list
            The vector field where a clustering task was run.
        alias: string
            Alias is used to name a cluster
        page_size: int
            Size of each page of results
        cursor: string
            Cursor to paginate the document retrieval
        page: int
            Page of the results
        include_vector: bool
            Include vectors in the search results
        similarity_metric: string
            Similarity Metric, choose from ['cosine', 'l1', 'l2', 'dp']
        """
        cluster_ids = [] if cluster_ids is None else cluster_ids

        return self.make_http_request(
            f"/datasets/{dataset_id}/cluster/centroids/documents",
            method="POST",
            parameters={
                # "dataset_id": dataset_id,
                "cluster_ids": cluster_ids,
                "vector_fields": vector_fields,
                "alias": alias,
                "page_size": page_size,
                "cursor": cursor,
                "page": page,
                "include_vector": include_vector,
                "similarity_metric": similarity_metric,
                "vector_field": "",
            },
        )

    list = documents

    def delete(
        self,
        dataset_id: str,
        centroid_id: str,
        alias: str,
        vector_fields: List[str],
        centroid_vector_fields: Optional[List[str]] = None,
        centroid_dataset_id: Optional[str] = None,
    ):
        """
        Delete centroids by dataset ID, vector field and alias

        Parameters
        ----------
        dataset_id : string
            Unique name of dataset
        vector_field: string
            The vector field where a clustering task was run.
        alias: string
            Alias is used to name a cluster

        """
        parameters = {
            "dataset_id": dataset_id,
            "vector_fields": vector_fields,
            "alias": alias,
        }

        if centroid_vector_fields is not None:
            parameters.update({"centroid_vector_fields": centroid_vector_fields})

        if centroid_dataset_id is not None:
            parameters.update({"centroid_dataset_id": centroid_dataset_id})

        return self.make_http_request(
            f"/datasets/{dataset_id}/cluster/centroids/{centroid_id}/delete",
            method="POST",
            parameters=parameters,
        )



"""Code for migrating datasets
"""
import math

from typing import Any, Dict, List, Optional

from tqdm.auto import tqdm


def migrate_dataset(
    old_token: str,
    new_token: str,
    dataset_id: str,
    new_dataset_id: Optional[str] = None,
    chunksize: int = 100,
    filters: list = None,
    show_progress_bar: bool = True,
):
    """
    Migrate dataset

    Args
    ---------
        old_token (str): _description_
        new_token (str): _description_
        dataset_id (str): _description_
        new_dataset_id (Optional[str], optional): _description_. Defaults to None.
        chunksize (int, optional): _description_. Defaults to 20.

    Example
    ---------

    .. code-block::

        from relevanceai.migration import migrate
        migrate_dataset(
            old_token="...",
            new_token="...",
            dataset_id="sample_dataset",
            new_dataset_id="new_sample_dataset")

    """
    from relevanceai import Client
    from relevanceai.utils.logger import FileLogger

    filters = filters if filters is not None else []

    if new_dataset_id is None:
        new_dataset_id = dataset_id

    old_client = Client(token=old_token)
    new_client = Client(token=new_token)

    old_dataset = old_client.Dataset(dataset_id)
    new_dataset = new_client.Dataset(dataset_id)

    after_id = None

    number_of_documents = old_client.get_number_of_documents(
        dataset_id=dataset_id, filters=filters
    )
    iterations_required = math.ceil(number_of_documents / chunksize)
    bar = tqdm(range(iterations_required), disable=(not show_progress_bar))
    overflow: List[Dict[str, Any]] = []

    with FileLogger():
        while True:
            res = old_dataset.get_documents(
                number_of_documents=chunksize,
                filters=filters,
                after_id=after_id,
            )
            documents = res["documents"]
            after_id = res["after_id"]

            if not documents:
                break

            for document in documents:
                # backwards compatibility for clusters
                if "_clusters_" in document:
                    document["_cluster_"] = document.pop("_clusters_")

            pool = documents + overflow
            n_batches = int(len(pool) / chunksize)

            for i in range(n_batches):
                batch_start = i * chunksize
                batch_end = (i + 1) * chunksize

                batch = pool[batch_start:batch_end]
                res = new_dataset.insert_documents(batch)

                failed_documents = res["failed_documents"]

                failed_ids = set(map(lambda x: x["_id"], failed_documents))
                failed_documents = [
                    document for document in documents if document["_id"] in failed_ids
                ]
                overflow += failed_documents

            bar.update(1)

        if overflow:
            new_dataset.upsert_documents(overflow)
            bar.update(1)

    tqdm.write("Finished migrating.")



"""Add Sentiment to your dataset
"""

# Running a function across each subcluster
import numpy as np
import csv
from typing import Optional
from urllib.request import urlopen
from relevanceai.constants.errors import MissingPackageError
from relevanceai.operations_new.transform_base import TransformBase


class SentimentTransform(TransformBase):
    def __init__(
        self,
        text_fields: list,
        model_name: str = "cardiffnlp/twitter-roberta-base-sentiment",
        highlight: bool = False,
        positive_sentiment_name: str = "positive",
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
        output_fields: list = None,
        sensitivity: float = 0,
        device: int = None,
        **kwargs,
    ):
        """
        Sentiment Ops.

        Parameters
        -------------

        model_name: str
            The name of the model
        sensitivity: float
            How confident it is about being `neutral`. If you are dealing with news sources,
            you probably want less sensitivity

        """
        self.model_name = model_name
        self.text_fields = text_fields
        self.highlight = highlight
        self.positive_sentiment_name = positive_sentiment_name
        self.max_number_of_shap_documents = max_number_of_shap_documents
        self.min_abs_score = min_abs_score
        self.output_fields = output_fields
        self.sensitivity = sensitivity
        self.device = self.get_transformers_device(device)

        import transformers

        # Tries to set it to set it on GPU
        self._classifier = transformers.pipeline(
            return_all_scores=True, model=self.model_name, device=self.device
        )

        for k, v in kwargs.items():
            setattr(self, k, v)

    def preprocess(self, text: str):
        new_text = []
        for t in text.split(" "):
            t = "@user" if t.startswith("@") and len(t) > 1 else t
            t = "http" if t.startswith("http") else t
            new_text.append(t)
        return " ".join(new_text)

    @property
    def classifier(self):
        return self._classifier

    # def _get_model(self):
    #     try:
    #         import transformers
    #     except ModuleNotFoundError:
    #         print(
    #             "Need to install transformers by running `pip install -q transformers`."
    #         )
    #     self.classifier = transformers.pipeline(
    #         "sentiment-analysis",
    #         return_all_scores=True,
    #         model="cardiffnlp/twitter-roberta-base-sentiment",
    #     )

    def _get_label_mapping(self, task: str):
        # Note: this is specific to the current model
        labels = []
        mapping_link = f"https://raw.githubusercontent.com/cardiffnlp/tweeteval/main/datasets/{task}/mapping.txt"
        with urlopen(mapping_link) as f:
            html = f.read().decode("utf-8").split("\n")
            csvreader = csv.reader(html, delimiter="\t")
        labels = [row[1] for row in csvreader if len(row) > 1]
        return labels

    @property
    def label_mapping(self):
        return {"LABEL_0": "negative", "LABEL_1": "neutral", "LABEL_2": "positive"}

    def analyze_sentiment(
        self,
        text,
        highlight: bool = False,
        positive_sentiment_name: str = "positive",
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
    ):
        if text is None:
            return None
        labels = self.classifier([str(text)], truncation=True, max_length=512)
        ind_max = np.argmax([l["score"] for l in labels[0]])
        sentiment = labels[0][ind_max]["label"]
        max_score = labels[0][ind_max]["score"]
        sentiment = self.label_mapping.get(sentiment, sentiment)
        if sentiment.lower() == "neutral" and max_score > self.sensitivity:
            overall_sentiment = 1e-5
        elif sentiment.lower() == "neutral":
            # get the next highest score
            new_labels = labels[0][:ind_max] + labels[0][(ind_max + 1) :]
            new_ind_max = np.argmax([l["score"] for l in new_labels])
            new_max_score = new_labels[new_ind_max]["score"]
            new_sentiment = new_labels[new_ind_max]["label"]
            new_sentiment = self.label_mapping.get(new_sentiment, new_sentiment)
            overall_sentiment = self._calculate_overall_sentiment(
                new_max_score, new_sentiment
            )

        else:
            overall_sentiment = self._calculate_overall_sentiment(max_score, sentiment)
        # Adjust to avoid bug
        if overall_sentiment == 0:
            overall_sentiment = 1e-5
        if not highlight:
            return {
                "sentiment": sentiment,
                "overall_sentiment_score": overall_sentiment,
            }
        shap_documents = self.get_shap_values(
            text,
            sentiment_ind=ind_max,
            max_number_of_shap_documents=max_number_of_shap_documents,
            min_abs_score=min_abs_score,
        )
        return {
            "sentiment": sentiment,
            "score": max_score,
            "overall_sentiment": overall_sentiment,
            "highlight_chunk_": shap_documents,
        }

    def _calculate_overall_sentiment(self, score: float, sentiment: str):
        if sentiment.lower().strip() == self.positive_sentiment_name:
            return score
        else:
            return -score

    @property
    def explainer(self):
        if hasattr(self, "_explainer"):
            return self._explainer
        else:
            try:
                import shap
            except ModuleNotFoundError:
                raise MissingPackageError("shap")
            self._explainer = shap.Explainer(self.classifier)
            return self._explainer

    def get_shap_values(
        self,
        text: str,
        sentiment_ind: int = 2,
        max_number_of_shap_documents: Optional[int] = None,
        min_abs_score: float = 0.1,
    ):
        """Get SHAP values"""
        shap_values = self.explainer([text])
        cohorts = {"": shap_values}
        cohort_labels = list(cohorts.keys())
        cohort_exps = list(cohorts.values())
        features = cohort_exps[0].data
        feature_names = cohort_exps[0].feature_names
        values = np.array([cohort_exps[i].values for i in range(len(cohort_exps))])
        shap_docs = [
            {"text": v, "score": f}
            for f, v in zip(
                [x[sentiment_ind] for x in values[0][0].tolist()], feature_names[0]
            )
        ]
        if max_number_of_shap_documents is not None:
            sorted_scores = sorted(shap_docs, key=lambda x: x["score"], reverse=True)
        else:
            sorted_scores = sorted(shap_docs, key=lambda x: x["score"], reverse=True)[
                :max_number_of_shap_documents
            ]
        return [d for d in sorted_scores if abs(d["score"]) > min_abs_score]

    @property
    def name(self):
        return "sentiment"

    def _get_output_field(self, text_field):
        norm_name = self.model_name.replace("/", "-")
        return f"_sentiment_.{text_field}.{norm_name}"

    def transform(self, documents):
        # For each document, update the field
        sentiment_docs = [{"_id": d["_id"]} for d in documents]
        for i, t in enumerate(self.text_fields):
            if self.output_fields is not None:
                output_field = self.output_fields[i]
            else:
                output_field = self._get_output_field(t)
            sentiments = [
                self.analyze_sentiment(
                    self.get_field(t, doc, missing_treatment="return_empty_string"),
                    highlight=self.highlight,
                    max_number_of_shap_documents=self.max_number_of_shap_documents,
                    min_abs_score=self.min_abs_score,
                )
                for doc in documents
            ]
            self.set_field_across_documents(output_field, sentiments, sentiment_docs)
        return sentiment_docs

    # def analyze_sentiment(self, text, highlight:bool= True):
    #     try:
    #         from scipy.special import softmax
    #     except ModuleNotFoundError:
    #         print("Need to install scipy")
    #     if not hasattr(self, "model"):
    #         self._get_model()
    #     text = self.preprocess(text)
    #     encoded_input = self.tokenizer(text, return_tensors="pt")
    #     output = self.model(**encoded_input)
    #     scores = output[0][0].detach().numpy()
    #     scores = softmax(scores)
    #     ranking = np.argsort(scores)
    #     ranking = ranking[::-1]
    #     sentiment = self.label_mapping[ranking[0]]
    #     score = np.round(float(scores[ranking[0]]), 4)
    #     return {
    #         "sentiment": sentiment,
    #         "score": np.round(float(scores[ranking[0]]), 4),
    #         "overall_sentiment_score": score if sentiment == "positive" else -score,
    #     }
